/*
 * ipclassify.c
 *
 *  Created on: Apr 19, 2011
 *      Author: tony
 */

/*
    License

    This file is part of ipclassify.

    ipclassify is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    ipclassify is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with ipclassify.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <signal.h>
#include <basic.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <syslog.h>
#include <pthread.h>
#include <regex.h>
#include <time.h>
#include "parser.h"
#include "config.h"




/* ----- Globals ------------------------------------------------------ */

bool debug = 0;
bool foreground = FALSE;
bool reReadConf = TRUE;
bool dumpStats = FALSE;
bool resetStats = FALSE;

char VERSION[] = "1.0";
char *pidfile = "/var/run/ipclassify.pid";
char *logfile = "/var/log/ipclassify.log";
char *statsfile = "/var/log/ipclassify.stats";

queue_t *threadlist = NULL;

pthread_mutex_t logfile_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t debug_mutex   = PTHREAD_MUTEX_INITIALIZER;

struct nfq_handle *h = NULL;

/* ----- Functions ---------------------------------------------------- */


char *
ip2str(in_addr_t ip)
{
  static char buf[2][ sizeof("aaa.bbb.ccc.ddd") ];
  static short int index=0;

  ip = ntohl(ip);

  sprintf(buf[index],"%d.%d.%d.%d",
      (ip >> 24) & 0xff,
      (ip >> 16) & 0xff,
      (ip >> 8) & 0xff,
      (ip) & 0xff);

  if (index) {
    index=0;
    return buf[1];
  }
  else return buf[index++];
}



void
Terminate(int final) {
	queue_t *qptr = threadlist;
	char mstr[DEFBUFSIZE];

	/* unlink the pid file */
	unlink(pidfile);

	log2file("Terminate(): Releasing queues...");
	debugMsg("Terminate(): Releasing queues...",1);
	while(qptr != NULL) {
		pthread_cancel(qptr->thread);
		pthread_rwlock_destroy(&qptr->rwlock);
		if(qptr->rbt != NULL) cp_rbtree_destroy(qptr->rbt);
		if(qptr->whitelist != NULL) cp_rbtree_destroy(qptr->whitelist);
		qptr = (queue_t *)qptr->next;
		free(threadlist);
		threadlist = qptr;
	}

	log2file("Terminate(): Unbinding netlink handle...");
	debugMsg("Terminate(): Unbinding netlink handle...",1);
	if(h != NULL) {
		nfq_close(h);
		h = NULL;
	}

	syslog(LOG_INFO, "Terminate().");
	log2file("Terminating...");

	/* Now we exit */
	exit(final);
}

void
log2stats(char *logstr) {
	FILE *file = NULL;
	char mstr[DEFBUFSIZE];

	file = fopen(statsfile, "a");
	if(!file) {
		snprintf((char *)mstr, DEFBUFSIZE, "log2stats(): %s: exiting, cannot write file...",
				logfile);
		syslog(LOG_ERR, mstr);
	    Terminate(EXIT_FAILURE);
	  }

	flockfile(file);
	fprintf(file, "%s\n", logstr);
	fflush(file);
	funlockfile(file);

	fclose(file);
	return;
}


void
log2file(char *logstr) {
	FILE *file = NULL;
	char mstr[DEFBUFSIZE];
	time_t t;
	struct tm *tm;
	char datestring[DEFBUFSIZE];

	t = time(NULL);
	tm = localtime(&t);
	if(tm == NULL) {
		snprintf((char *)mstr, DEFBUFSIZE, "log2file(): localtime(): failure");
		syslog(LOG_ERR, mstr);
	    Terminate(EXIT_FAILURE);
	}
	strftime (datestring, sizeof(datestring), "%c", tm);

	pthread_mutex_lock(&logfile_mutex);
	file = fopen(logfile, "a");
	if(!file) {
		snprintf((char *)mstr, DEFBUFSIZE, "log2file(): %s: exiting, cannot write log...",
				logfile);
		syslog(LOG_ERR, mstr);
	    Terminate(EXIT_FAILURE);
	  }
	flockfile(file);
	fprintf(file, "%d: %s: %s\n", (int)getpid(), datestring, logstr);
	fflush(file);
	funlockfile(file);
	pthread_mutex_unlock(&logfile_mutex);

	fclose(file);
	return;
}


void
debugMsg(char *str, int level)
{
	if(foreground && (debug >= level)) {
		pthread_mutex_lock(&debug_mutex);
		fprintf(stdout,"\tDEBUG: thread 0x%x: %s\n",
			 pthread_self(), str);
		fflush(stdout);
		pthread_mutex_unlock(&debug_mutex);
	}
}


void
SigHandle(int sig)
{
	char mstr[DEFBUFSIZE];

	snprintf((char *)mstr, DEFBUFSIZE, "Caught Signal %d",
			sig);
	syslog(LOG_INFO,mstr);
	log2file(mstr);
	debugMsg(mstr,1);
	if(sig == SIGHUP) {
		reReadConf = TRUE;
		return;
	} else if(sig == SIGUSR1) {
		dumpStats = TRUE;
		return;
	} else if(sig == SIGUSR2) {
		resetStats = TRUE;
		return;
	} else if(sig == SIGTERM) {
		snprintf((char *)mstr, DEFBUFSIZE, " normal TERM, Dumping final stats.");
		log2file(mstr);
		debugMsg(mstr,1);
		log2stats("|---------------------------------------------------------------------------|");
		log2stats("\t!! Got TERM, Final Stats !!");
		RBTDumpStats(threadlist);
	} else if(sig == SIGABRT) { /*die like a dog */ unlink(pidfile); exit(EXIT_FAILURE); }

	Terminate(sig);
}

void
queueStats(void) {
	queue_t *qptr = threadlist;
	char mstr[DEFBUFSIZE];

	while(qptr != NULL) {
		if(qptr->rbt != NULL) {
			snprintf((char *)mstr,DEFBUFSIZE,"Final load stats: queue %d has %d nodes",
					qptr->queue,cp_rbtree_count(qptr->rbt));
			log2file(mstr);
			debugMsg(mstr,1);
		}
		if(qptr->whitelist != NULL) {
			snprintf((char *)mstr,DEFBUFSIZE,"Final load stats: queue whitelist %d has %d nodes",
					qptr->queue,cp_rbtree_count(qptr->whitelist));
			log2file(mstr);
			debugMsg(mstr,1);
		}
		qptr = (queue_t *)qptr->next;
	}
	return;
}

int
checkPID()
{
  FILE *f;
  char buffer[128];

  if(debug) {
    fprintf(stdout,"\tcheckPID()\n");
  }

  if ((f = fopen(pidfile, "r")) != NULL) {
    if(fgets(buffer, sizeof(buffer), f) == NULL){
      syslog(LOG_INFO,"-- ERROR -- Cannot Read pidfile \"%s\"",pidfile);
      exit(-1);
    }
    syslog(LOG_INFO,"-- ERROR -- Already running pid \"%s\" ? (remove pidfile \"%s\").",buffer,pidfile);
    fprintf(stderr,"\n\t!! ERROR: Already running, pid \"%s\"\n",
     buffer);
    fprintf(stderr,"\t   Note: Remove pidfile \"%s\" to override !\n\n",
     pidfile);
    fclose(f);
    exit(-1);
  }
  f = fopen(pidfile, "w");
  if (f) {
    fprintf(f, "%d", getpid());
    fclose(f);
  } else {
    syslog(LOG_INFO,"Error creating pidfile %s: %s", pidfile,
         strerror(errno));
  }
  return(SUCCESS);
}

void
emptyThreadList(void) {
  queue_t *tptr = NULL, *cptr = NULL;

  if(threadlist == NULL) {return;}
  tptr = threadlist;
  while(tptr != NULL) {
    cptr = tptr;
    pthread_cancel(tptr->thread);
	pthread_rwlock_destroy(&tptr->rwlock);
    if(tptr->rbt != NULL) {cp_rbtree_destroy(tptr->rbt);}
    if(tptr->whitelist != NULL) {cp_rbtree_destroy(tptr->whitelist);}
    tptr = (queue_t *)cptr->next;
    free(cptr);
  }
  threadlist = NULL;
  if(h != NULL) {
	  nfq_close(h);
	  nfq_unbind_pf(h, AF_INET); /* BUG ?: may need to remove this ? */
	  h = NULL;
  }
}

void
usage(void) {
  fprintf(stdout,"\n\n");
  fprintf(stdout,"\tipclassify [-f] [-c file] [-s statsfile] [-p pidfile] [-l logfile] [-d #] [-v|-V]\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"\t ver: %s\n",VERSION);
  fprintf(stdout,"\n");
  fprintf(stdout,"\twhere:\n");
  fprintf(stdout,"\t\t-d\tdebugging level (default none).\n");
  fprintf(stdout,"\t\t\t  (implies forground mode)\n");
  fprintf(stdout,"\t\t-V|-v\tPrint Version, then exit.\n");
  fprintf(stdout,"\t\t-c\tParse config file \"file\" (default /etc/ipclassify.conf)\n");
  fprintf(stdout,"\t\t-s\tSTATS file \"statsfile\" (default %s)\n", statsfile);
  fprintf(stdout,"\t\t-p\tPID file \"pidfile\" (default %s)\n", pidfile);
  fprintf(stdout,"\t\t-l\tlog file \"logfile\" (default %s)\n", logfile);
  fprintf(stdout,"\t\t-f\tRun in foreground (default daemon).\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"\tSignals:\n");
  fprintf(stdout,"\t\tHUP - ReRead configuration file.\n");
  fprintf(stdout,"\t\t\t(Resets Stats to Zero)\n");
  fprintf(stdout,"\t\tUSR1 - Write current Stats to Stats file.\n");
  fprintf(stdout,"\t\t\t(default %s)\n",statsfile);
  fprintf(stdout,"\t\tUSR2 - Reset all Stats.\n");
  fprintf(stdout,"\n\n");
  exit(EXIT_SUCCESS);
}

/* find the queue on the threadlist, return NULL if not there */
queue_t *
findQueue(int queue){
	queue_t *tmpptr = NULL;

	if(threadlist == NULL) {return(NULL);}
	tmpptr = threadlist;
	while(tmpptr != NULL) {
		if(tmpptr->queue == queue) {return(tmpptr);}
		tmpptr = (queue_t *)tmpptr->next;
	}
	return(NULL);
}


/* Create new queue int and return pointer to new node */
queue_t *
addQueue(int queue) {
	queue_t *tmpptr = NULL;

	/* if this is the first */
	if(threadlist == NULL){
		/* add a new node */
		if((threadlist = calloc(sizeof(queue_t),1)) == NULL) {
			syslog(LOG_ERR,"calloc() failure in addQueue()");
			log2file("calloc() failure in addQueue()");
			debugMsg("calloc() failure in addQueue()",1);
			Terminate(EXIT_FAILURE);
		}
		threadlist->queue = queue;
		threadlist->next = NULL;
		threadlist->rbt = NULL;
		threadlist->whitelist = NULL;
		threadlist->qh = NULL;
		threadlist->ttl_packets = threadlist->ttl_rejected = 0;
		threadlist->acceptMark = threadlist->rejectMark = 0;
		if(pthread_rwlock_init(&threadlist->rwlock,NULL) != 0) {
			syslog(LOG_ERR,"pthread_rwlock_init() failure in addQueue()");
			log2file("pthread_rwlock_init() failure in addQueue()");
			debugMsg("pthread_rwlock_init() failure in addQueue()",1);
			Terminate(EXIT_FAILURE);
		}
		return(threadlist);
	}

	/* goto the end of the threadlist */
	tmpptr = threadlist;

	while(tmpptr->next != NULL) {tmpptr = (queue_t *)tmpptr->next;}

	/* add a new node */
	if((tmpptr->next = calloc(sizeof(queue_t),1)) == NULL) {
		syslog(LOG_ERR,"calloc() failure in addQueue()");
		log2file("calloc() failure in addQueue()");
		debugMsg("calloc() failure in addQueue()",1);
		Terminate(EXIT_FAILURE);
	}
	/* change to the newly add node and initialize everything */
	tmpptr = (queue_t *)tmpptr->next;
	tmpptr->queue = queue;
	tmpptr->next = NULL;
	tmpptr->rbt = NULL;
	tmpptr->whitelist = NULL;
	tmpptr->qh = NULL;
	tmpptr->ttl_packets = tmpptr->ttl_rejected = 0;
	tmpptr->acceptMark = tmpptr->rejectMark = 0;
	if(pthread_rwlock_init(&tmpptr->rwlock,NULL) != 0) {
		syslog(LOG_ERR,"pthread_rwlock_init() failure in addQueue()");
		log2file("pthread_rwlock_init() failure in addQueue()");
		debugMsg("pthread_rwlock_init() failure in addQueue()",1);
		Terminate(EXIT_FAILURE);
	}

	return(tmpptr);
}


void
ranged_insert(int queue, char *name, char *ipmin, char *ipmax, int whitelist) {
	char mstr[DEFBUFSIZE];
	queue_t *node = NULL;

	snprintf((char *)mstr, DEFBUFSIZE, "ranged_insert(%d, %s, %s, %s)",
			queue, name, ipmin, ipmax);
	/* OK, this is redundant, but saves a function call overhead when not in foreground */
	if(debug) debugMsg(mstr,2);

	node = findQueue(queue);
	if(node == NULL) {
		if((node = addQueue(queue)) == NULL) {
			snprintf((char *)mstr, DEFBUFSIZE, "ranged_insert(): addQueue(%d): failed",
					queue);
			syslog(LOG_ERR,mstr);
			debugMsg(mstr,1);
			Terminate(EXIT_FAILURE);
		}
	}
	addToRBT(node,name,ipmin,ipmax, whitelist);

	if(node->rbt != NULL) {
		snprintf((char *)mstr, DEFBUFSIZE, "ranged_insert(): nodes in tree %d: %lu",
			queue, cp_rbtree_count(node->rbt));
		/* OK, this is redundant, but saves a function call overhead when not in foreground */
		if(debug) debugMsg(mstr,2);
	}
	if(node->whitelist != NULL) {
		snprintf((char *)mstr, DEFBUFSIZE, "ranged_insert(): nodes in whitelist %d: %lu",
			queue, cp_rbtree_count(node->whitelist));
		/* OK, this is redundant, but saves a function call overhead when not in foreground */
		if(debug) debugMsg(mstr,2);
	}


	return;
}

void
loadfile(int queue, char *infile, int whitelist) {
	FILE *file = NULL;
	char mstr[DEFBUFSIZE];
	regex_t regmain;
	regmatch_t matches[4];
	char *line = NULL;
	size_t len = 0;
	int ntot = 0,count = 0, i = 0;

	snprintf((char *)mstr, DEFBUFSIZE, "loadfile(\"%d\",\"%s\",\"%d\")",
			queue,infile,whitelist);
	/* OK, this is redundant, but saves a function call overhead when not in foreground */
	if(debug) debugMsg(mstr,1);
	snprintf((char *)mstr, DEFBUFSIZE, "load file %s for queue %d",
			infile, queue);
	log2file(mstr);

	regcomp(&regmain, "^(.*)[:]([0-9.]*)[-]([0-9.]*)$", REG_EXTENDED);

	file = fopen(infile,"r");
	if(!file) {
		snprintf((char *)mstr, DEFBUFSIZE, "loadfile(\"%d\",\"%s\"): Failed to open file",
				queue,infile);
		  log2file(mstr);
		  regfree(&regmain);
		  return;
	}

	while ( (count=getline(&line,&len,file)) != -1 ) {
		if ( line[0] == '#' )		//comment line, skip
			continue;
		for(i=count-1; i>=0; i--) {
			if ((line[i] == '\r') || (line[i] == '\n') || (line[i] == ' ')) {
				line[i] = 0;
			} else {
				break;
			}
		}

		if (strlen(line) == 0)
			continue;

		if (!regexec(&regmain, line, 4, matches, 0)) {
			line[matches[1].rm_eo] = 0;
			line[matches[2].rm_eo] = 0;
			line[matches[3].rm_eo] = 0;

			ranged_insert(queue, line+matches[1].rm_so,
				      line+matches[2].rm_so,
				      line+matches[3].rm_so,
				      whitelist);

			ntot++;
			snprintf((char *)mstr, DEFBUFSIZE, "loadfile(): Ranges read: %d", ntot);
			/* OK, this is redundant, but saves a function call overhead when not in foreground */
			if(debug) debugMsg(mstr,2);
		} else {
			snprintf((char *)mstr, DEFBUFSIZE,"loadfile(): Short line %s, skipping it...", line);
			log2file(mstr);
		}
	}

	if (line)
		free(line);

	fclose(file);
	regfree(&regmain);

	snprintf((char *)mstr, DEFBUFSIZE, "loadfile(): Ranges loaded: %d", ntot);
	debugMsg(mstr,1);
	log2file(mstr);

	return;
}


void
setMark(int accept, int queue, int mark) {
	char mstr[DEFBUFSIZE];
	queue_t *node = NULL;

	node = findQueue(queue);
	if(node == NULL) {
		if((node = addQueue(queue)) == NULL) {
			snprintf((char *)mstr, DEFBUFSIZE, "setMark(): addQueue(%d): failed",
					queue);
			syslog(LOG_ERR,mstr);
			debugMsg(mstr,1);
			Terminate(EXIT_FAILURE);
		}
	}

	if(accept) {
		node->acceptMark = mark;
		snprintf((char *)mstr, DEFBUFSIZE, "setMark(): Queue %d accept mark set to %d", queue,mark);
	} else {
		node->rejectMark = mark;
		snprintf((char *)mstr, DEFBUFSIZE, "setMark(): Queue %d reject mark set to %d", queue,mark);
	}

	debugMsg(mstr,1);
	log2file(mstr);

	return;
}



/*
 * Largely from the following source:
 * http://netfilter.org/projects/libnetfilter_queue/doxygen/nfqnl__test_8c_source.html
 *
 */

static int
nfqueue_cb(struct nfq_q_handle *qh, struct nfgenmsg *nfmsg,
            struct nfq_data *nfa, void *data)
{
	char mstr[DEFBUFSIZE];
	int id = 0, status = 0;
	struct nfqnl_msg_packet_hdr *ph = NULL;
	unsigned char *payload = NULL;
	recType tmprec = {NULL,0,0};
	queue_t *queue = NULL;

	pthread_testcancel();
	queue = (queue_t *)data;
	ph = nfq_get_msg_packet_hdr(nfa);
	if (ph) {
		id = ntohl(ph->packet_id);
	    nfq_get_payload(nfa, &payload);

	    /* increment the number of packets seen by this queue */
	    pthread_rwlock_wrlock(&queue->rwlock);
	    queue->ttl_packets++;
	    pthread_rwlock_unlock(&queue->rwlock);

        switch (ph->hook) {
	      case NF_IP_LOCAL_IN:
	    	  if(findInRBT(queue,ntohl(SRC_ADDR(payload)),&tmprec) == SUCCESS) {
	              // we drop the packet instead of rejecting
	              // we don't want the other host to know we are alive
	    		  pthread_rwlock_wrlock(&queue->rwlock);
	    		  queue->ttl_rejected++;
	    		  pthread_rwlock_unlock(&queue->rwlock);
	              status=nfq_set_verdict2(qh, id, NF_DROP, queue->rejectMark, 0, NULL);
	              if(status < 0) {
	            	  snprintf((char *)mstr,DEFBUFSIZE,"Queue: %d: nfq_set_verdict():error: %s, hits: %d, SRC: %s",
	            			  queue->queue,tmprec.blockname,tmprec.hits,ip2str(SRC_ADDR(payload)));
	            	  log2file(mstr);
	            	  debugMsg(mstr,1);
	            	  syslog(LOG_INFO,"Queue: %d: nfq_set_verdict():error: %s, hits: %d, SRC: %s",
	            			  queue->queue,tmprec.blockname,tmprec.hits,ip2str(SRC_ADDR(payload)));
	            	  free(tmprec.blockname);
	            	  return(1);
	              } else {
	            	  snprintf((char *)mstr,DEFBUFSIZE,"Blocked IN: Queue: %d: %s, hits: %d, SRC: %s",
	            			  queue->queue,tmprec.blockname,tmprec.hits,ip2str(SRC_ADDR(payload)));
	            	  log2file(mstr);
	            	  debugMsg(mstr,1);
	            	  free(tmprec.blockname);
	              }
	    	  } else {
	    		  // NF_ACCEPT the packet
	              status = nfq_set_verdict2(qh, id, NF_ACCEPT, queue->acceptMark, 0, NULL);
	              if(status < 0) {
	            	  snprintf((char *)mstr,DEFBUFSIZE,"Queue: %d: nfq_set_verdict():error: ",queue->queue);
	            	  log2file(mstr);
	            	  debugMsg(mstr,1);
	            	  syslog(LOG_INFO,"Queue: %d: nfq_set_verdict():error: %s, hits: %d, SRC: %s, DEST: %s",queue->queue);
	            	  return(1);
	              }
	          }
	    	  break;
	      case NF_IP_LOCAL_OUT:
	    	  if(findInRBT(queue,ntohl(DEST_ADDR(payload)),&tmprec) == SUCCESS) {
	              // we drop the packet instead of rejecting
	              // we don't want the other host to know we are alive
	    		  pthread_rwlock_wrlock(&queue->rwlock);
	    		  queue->ttl_rejected++;
	    		  pthread_rwlock_unlock(&queue->rwlock);
	              status=nfq_set_verdict2(qh, id, NF_DROP, queue->rejectMark, 0, NULL);
	              if(status < 0) {
	            	  snprintf((char *)mstr,DEFBUFSIZE,"Queue: %d: nfq_set_verdict():error: %s, hits: %d, DEST: %s",
	            			  queue->queue,tmprec.blockname,tmprec.hits,ip2str(DEST_ADDR(payload)));
	            	  log2file(mstr);
	            	  debugMsg(mstr,1);
	            	  syslog(LOG_INFO,"Queue: %d: nfq_set_verdict():error: %s, hits: %d, DEST: %s",
	            			  queue->queue,tmprec.blockname,tmprec.hits,ip2str(DEST_ADDR(payload)));
	            	  free(tmprec.blockname);
	            	  return(1);
	              } else {
	            	  snprintf((char *)mstr,DEFBUFSIZE,"Blocked OUT: Queue: %d: %s, hits: %d, DEST: %s",
	            			  queue->queue,tmprec.blockname,tmprec.hits,ip2str(DEST_ADDR(payload)));
	            	  log2file(mstr);
	            	  debugMsg(mstr,1);
	            	  free(tmprec.blockname);
	              }
	    	  } else {
	    		  // NF_ACCEPT the packet
	              status = nfq_set_verdict2(qh, id, NF_ACCEPT, queue->acceptMark, 0, NULL);
	              if(status < 0) {
	            	  snprintf((char *)mstr,DEFBUFSIZE,"Queue: %d: nfq_set_verdict():error: ",queue->queue);
	            	  log2file(mstr);
	            	  debugMsg(mstr,1);
	            	  syslog(LOG_INFO,"Queue: %d: nfq_set_verdict():error: %s, hits: %d, SRC: %s, DEST: %s",queue->queue);
	            	  return(1);
	              }
	          }
	    	  break;
	      case NF_IP_FORWARD:
	    	  if((findInRBT(queue,ntohl(SRC_ADDR(payload)),&tmprec) == SUCCESS) ||
	    			  (findInRBT(queue,ntohl(DEST_ADDR(payload)),&tmprec) == SUCCESS)){
	              // we drop the packet instead of rejecting
	              // we don't want the other host to know we are alive
	    		  pthread_rwlock_wrlock(&queue->rwlock);
	    		  queue->ttl_rejected++;
	    		  pthread_rwlock_unlock(&queue->rwlock);
	              status=nfq_set_verdict2(qh, id, NF_DROP, queue->rejectMark, 0, NULL);
	              if(status < 0) {
	            	  snprintf((char *)mstr,DEFBUFSIZE,"Queue: %d: nfq_set_verdict():error: %s, hits: %d, SRC: %s, DEST: %s",
	            			  queue->queue,tmprec.blockname,tmprec.hits,ip2str(SRC_ADDR(payload)),ip2str(DEST_ADDR(payload)));
	            	  log2file(mstr);
	            	  debugMsg(mstr,1);
	            	  syslog(LOG_INFO,"Queue: %d: nfq_set_verdict():error: %s, hits: %d, SRC: %s, DEST: %s",
	            			  queue->queue,tmprec.blockname,tmprec.hits,ip2str(DEST_ADDR(payload)),ip2str(DEST_ADDR(payload)));
	            	  free(tmprec.blockname);
	            	  return(1);
	              } else {
	            	  snprintf((char *)mstr,DEFBUFSIZE,"Blocked FORWARD: Queue: %d: %s, hits: %d, SRC: %s, DEST: %s",
	            			  queue->queue,tmprec.blockname,tmprec.hits,ip2str(SRC_ADDR(payload)),ip2str(DEST_ADDR(payload)));
	            	  log2file(mstr);
	            	  debugMsg(mstr,1);
	            	  free(tmprec.blockname);
	              }
	    	  } else {
	    		  // NF_ACCEPT the packet
	              status = nfq_set_verdict2(qh, id, NF_ACCEPT, queue->acceptMark, 0, NULL);
	              if(status < 0) {
	            	  snprintf((char *)mstr,DEFBUFSIZE,"Queue: %d: nfq_set_verdict():error: ",queue->queue);
	            	  log2file(mstr);
	            	  debugMsg(mstr,1);
	            	  syslog(LOG_INFO,"Queue: %d: nfq_set_verdict():error: %s, hits: %d, SRC: %s, DEST: %s",queue->queue);
	            	  return(1);
	              }
	    	  }
	    	  break;
	      default:
	        log2file("Not NF_LOCAL_IN/OUT/FORWARD packet!");
	        debugMsg("Not NF_LOCAL_IN/OUT/FORWARD packet!",1);
	        break;
	    }
	  }
	  else {
	    log2file("NFQUEUE: can't get msg packet header.");
	    debugMsg("NFQUEUE: can't get msg packet header.",1);
	    return(1);    // from nfqueue source: 0 = ok, >0 = soft error, <0 hard error
	  }

	return(0);
}


/*
 * Largely from the following source:
 * http://netfilter.org/projects/libnetfilter_queue/doxygen/nfqnl__test_8c_source.html
 *
 */

void netlink_init(void) {
	char mstr[DEFBUFSIZE];

	snprintf((char *)mstr,DEFBUFSIZE,"Initializing nfnetlink..");
	log2file(mstr);
	debugMsg(mstr,1);

	if(h != NULL) {
		log2file("NFQ handle not NULL, must already be there");
		debugMsg("NFQ handle not NULL, must already be there",1);
		return;
	}

	h = nfq_open();
	if (!h) {
		log2file("Error during nfq_open()");
		debugMsg("Error during nfq_open()",1);
	    Terminate(-1);
	}

	if (nfq_unbind_pf(h, AF_INET) < 0) {
	    log2file("error during nfq_unbind_pf()");
	    debugMsg("error during nfq_unbind_pf()",1);
	}


	if (nfq_bind_pf(h, AF_INET) < 0) {
	    log2file("Error during nfq_bind_pf()");
	    debugMsg("Error during nfq_bind_pf()",1);
	    Terminate(-1);
	}
	return;
}

void *worker_thread_loop(void *argptr) {
	struct nfnl_handle *nh;
	int fd,rv;
	char buf[BUFSIZE];
	char mstr[DEFBUFSIZE];
	queue_t *queue = NULL;

	queue = (queue_t *)argptr;

	snprintf((char *)mstr,DEFBUFSIZE,"worker thread created for queue '%hd'", queue->queue);
	log2file(mstr);
	debugMsg(mstr,1);

	snprintf((char *)mstr,DEFBUFSIZE,"NFQUEUE: binding to queue '%hd'", queue->queue);
	log2file(mstr);
	debugMsg(mstr,1);
	queue->qh = nfq_create_queue(h,  queue->queue, &nfqueue_cb, argptr);
	if (!queue->qh) {
		snprintf((char *)mstr,DEFBUFSIZE,"error during nfq_create_queue(): queue '%hd'",
				queue->queue);
	    log2file(mstr);
	    debugMsg(mstr,1);
	    Terminate(-1);
	}

	pthread_cleanup_push((void *)nfq_destroy_queue, (void *)queue->qh);

	if (nfq_set_mode(queue->qh, NFQNL_COPY_PACKET, PAYLOADSIZE) < 0) {
	    log2file("can't set packet_copy mode");
	    debugMsg("can't set packet_copy mode",1);
	    Terminate(-1);
	}

	nh = nfq_nfnlh(h);
	fd = nfnl_fd(nh);

	while ((rv = recv(fd, buf, sizeof(buf), 0)) >= 0) {
		pthread_testcancel();
		nfq_handle_packet(h, buf, rv);
	}
	pthread_cleanup_pop(1);

	return(NULL);
}




int
main( int argc, char **argv) {
  char *cfile = NULL;
  FILE *file = NULL;
  extern FILE *yyin;
  int c, pid = 0;
  sigset_t sigs_to_block;
  queue_t *qptr = NULL;
  char mstr[DEFBUFSIZE];

  while((c = getopt(argc,argv,"hfVvc:s:p:l:d:")) != -1) {
    switch(c) {
    case 'c':
      cfile = optarg;
      break;
    case 's':
      statsfile = optarg;
      break;
   case 'p':
      pidfile = optarg;
      break;
    case 'l':
      logfile = optarg;
      break;
    case 'f':
      foreground = TRUE;
      break;
    case 'd':
      debug = (bool)atoi(optarg);
      /* any debug implies in the foreground */
      foreground = TRUE;
      break;
    case 'v':
    case 'V':
    	fprintf(stdout,"Version: %s\n", VERSION);
    	exit(EXIT_SUCCESS);
    	break;
    case 'h':
    default:
      usage();
      break;
    }
  }

   if(getuid() != 0) {
     fprintf(stderr,"\n\t!! ERROR: %s: must be root to run !!\n\n", argv[0]);
     exit(-1);
   }

  /* if no configuration file supplied on command line, use the default file */
  if(cfile == NULL) {
    cfile="/etc/ipclassify.conf";
  }

  if(debug) fprintf(stdout,"\n%s: Configuration file: %s\n",argv[0], cfile);

  /* if running in foreground do not fork process */
  if(foreground) {
	  fprintf(stdout,"\n%s: running in foreground, no fork()\n", argv[0]);
  }else {
	  pid = fork();
	  if(pid) exit(EXIT_SUCCESS);
  }

  /* if not in foreground close stdout and stderr
   *  and re-direct it to /dev/null.
   */
  if(!foreground) {
    /* STDIN */
    if(close(0) != 0) {
      perror("Cannot redirect stdin !\n");
      exit(-2);
    }
    if(open("/dev/zero",O_RDONLY) < 0) {
      perror("Cannot redirect stdin !\n");
      exit(-2);
    }
    /* STDOUT */
    if(close(1) != 0) {
      perror("Cannot redirect stdout !\n");
      exit(-2);
    }
    if(open("/dev/null",O_WRONLY) < 0) {
      perror("Cannot redirect stdout !\n");
      exit(-2);
    }
    /* STDERR */
    if(close(2) != 0) {
      perror("Cannot redirect stderr !\n");
      exit(-2);
    }
    if(open("/dev/null",O_WRONLY) < 0) {
      perror("Cannot redirect stderr !\n");
      exit(-2);
    }
  }

  openlog(argv[0], LOG_PID | LOG_NDELAY, LOG_LOCAL1);
  syslog(LOG_INFO, "Starting up.");
  log2file("Starting up.");

  /* after this we only exit calling the Terminate() function to clean up */
  (void)checkPID();

  debugMsg("Setting up signal handlers",1);
  /* install signal handlers */
  (void)signal(SIGHUP, SigHandle);
  (void)signal(SIGINT, SigHandle);
  (void)signal(SIGQUIT,SigHandle);
  (void)signal(SIGABRT,SigHandle);
  (void)signal(SIGKILL,SigHandle);
  (void)signal(SIGUSR1,SigHandle);
  (void)signal(SIGUSR2,SigHandle);
  (void)signal(SIGTERM,SigHandle);

  /* time to setup sigmask to block for all signals */
  c = sigemptyset(&sigs_to_block);
  if(c != 0) {
	  syslog(LOG_ERR, "sigemptyset() failure");
	  log2file("sigemptyset() failure");
	  debugMsg("sigemptyset() failure",1);
	  Terminate(EXIT_FAILURE);
  }

  c = sigaddset(&sigs_to_block, SIGUSR1);
  c += sigaddset(&sigs_to_block, SIGUSR2);
  c += sigaddset(&sigs_to_block, SIGHUP);
  c += sigaddset(&sigs_to_block, SIGINT);
  c += sigaddset(&sigs_to_block, SIGQUIT);
  c += sigaddset(&sigs_to_block, SIGABRT);
  c += sigaddset(&sigs_to_block, SIGKILL);
  c += sigaddset(&sigs_to_block, SIGTERM);
  if(c != 0) {
	  syslog(LOG_ERR, "sigaddset() failure");
	  log2file("sigaddset() failure");
	  debugMsg("sigaddset() failure",1);
	  Terminate(EXIT_FAILURE);
  }

  while(TRUE) {
	  if(reReadConf){
		  /* if there are any threads, cancel them */
		  if(threadlist != NULL) {
			  qptr = threadlist;
			  while(qptr != NULL){
				  snprintf((char *)mstr,DEFBUFSIZE,"Main: cancelling thread for queue %d",
						  qptr->queue);
				  log2file(mstr);
				  debugMsg(mstr,1);
				  pthread_cancel(qptr->thread);
				  qptr = (queue_t *)qptr->next;
			  }
			  /* Zero out the threadlist */
			  emptyThreadList();
		  }

		  /* Open the config file */
		  if(cfile == NULL) { /* BUG should never enter this if */
			  log2file("Come on Man, No config file ?!!");
			  debugMsg("Come on Man, No config file ?!!",1);
			  syslog(LOG_ERR,"Come on Man, No config file ?!!");
			  Terminate(EXIT_FAILURE);
		  }
		  file = fopen(cfile, "r");
		  if(!file) {
			snprintf((char *)mstr,DEFBUFSIZE,"Cannot open \"%s\" !!",cfile);
			log2file(mstr);
			debugMsg(mstr,1);
			syslog(LOG_ERR,mstr);
		    Terminate(EXIT_FAILURE);
		  }
		  flockfile(file);
		  yyin=file;

		  /* Parse the configuration file */
		  yyparse();

		  funlockfile(file);
		  fclose(file);
		  reReadConf = FALSE;

		  queueStats();
		  netlink_init();

		  pthread_sigmask(SIG_BLOCK, &sigs_to_block, NULL);
		  /* time to start the queue threads */
		  qptr = threadlist;
		  while(qptr != NULL) {
			  pthread_create(&qptr->thread,NULL,worker_thread_loop,qptr);
			  qptr = (queue_t *)qptr->next;
		  }
		  pthread_sigmask(SIG_UNBLOCK, &sigs_to_block, NULL);
	  }
	  if(dumpStats) {
		  debugMsg("Dumping to Stats File",1);
		  log2file("Dumping to Stats File");
		  RBTDumpStats(threadlist);
		  dumpStats = FALSE;
	  }
	  if(resetStats){
		  debugMsg("Resetting all Stats",1);
		  log2file("Resetting all Stats");
		  RBTResetStats(threadlist);
		  resetStats = FALSE;
	  }

	  /* it should be possible to eliminate this sleep() with some kind of mutex condition */
	  sleep(10);

  }

  log2file("Normal exit.");
  debugMsg("Normal exit.",1);
  syslog(LOG_INFO, "Normal exit.");
  Terminate(EXIT_SUCCESS);
  return(EXIT_SUCCESS); /* Never reach this statement, just keeps lint happy */
}
